(****************************************************************************

ZMIXER v.4 - very simple Cyrillic string cipher with two 16 char secret keys.

Key may be any ASCII string from 1 to 16 characters.

Toy cipher by (C) 2009 Alexander Myasnikov

Web: www.darksoftware.narod.ru

****************************************************************************)


{$R-}

unit zmixer;

interface

function Decrypt (s: string): string;
function Crypt (s: string): string;
procedure SetKey (s, s2: string);


implementation

const
  lowr = 192;

const
  highr = 255;

const
  modval = (highr - lowr) + 1;

var
  stable: array [lowr..highr] of byte;
  bstable: array [lowr..highr] of byte;
  stable2: array [lowr..highr] of byte;
  bstable2: array [lowr..highr] of byte;
  key: array [1..16] of byte;
  key2: array[1..16] of longword;
  iv: longword;


procedure SetKey (s, s2: string);
var
  i, x, y, vec, max: longword;
  n: byte;
  was: array [lowr..highr] of byte;
  a: array [1..16] of longword;
begin

  x := 1;

  s := copy(s, 1, 16);

  for I := lowr to highr do
    was[i] := 0;


  if length(s) < 16 then
    begin
    for I := 0 to 15 - length(s) do
      s := s + s[(i mod length(s)) + 1];
    end;


  s2 := copy(s2, 1, 16);

  if length(s2) < 16 then
    begin
    for I := 0 to 15 - length(s2) do
      s2 := s2 + s2[(i mod length(s2)) + 1];
    end;


  vec := 1;

  for I := 1 to 16 do
    begin
    a[i] := byte(s[I]);
    key[i] := (a[i] mod modval) + lowr;
    key2[i] := (byte(s2[i]) + i * 3) xor $FFFFFF;
    vec := vec + (a[i] * i);
    x := (x * (x + 1) shr 1) + a[i];
    end;

  iv := (vec mod modval) + lowr;


  for I := lowr to highr do
    begin

    repeat

      a[(x mod 16) + 1] := (a[1] + a[2]) + ((a[3] - a[4]) xor
        (a[5] shr (a[6] mod 8))) + (a[7] shl (a[8] mod 8)) +
        (a[6] xor a[8]) xor ((a[5] + x) xor (a[7] + i)) +
        (a[9] - a[10] + ((a[10] shl 6) xor (a[10] shr 11))) +
        (a[12] + a[13]) + (a[14] xor a[15]) + (a[16] xor (a[16] * 9)) +
        (a[1] + (a[1] xor x));

      n := (a[(x mod 16) + 1] mod modval) + lowr;
      Inc(x);


    until (was[n] = 0) and ((n <> i) or (i = highr));


    if (n = i) and (i = highr) then
      begin
      x := stable[highr - 1];
      stable[highr - 1] := n;
      bstable[n] := highr - 1;
      n := x;
      end;

    was[n] := 1;

    stable[i]  := n;
    bstable[n] := i;

    end;


  for I := lowr to highr do
    was[i] := 0;

  x := 1;

  for I := 1 to 16 do
    begin
    x := (x * (x + 1) shr 1) + a[i];
    end;

  for I := lowr to highr do
    begin

    repeat

      a[(x mod 16) + 1] := (a[1] + a[2]) + ((a[3] - a[4]) xor
        (a[5] shr (a[6] mod 8))) + (a[7] shl (a[8] mod 8)) +
        (a[6] xor a[8]) xor ((a[5] + x) xor (a[7] + i)) +
        (a[9] - a[10] + ((a[10] shl 6) xor (a[10] shr 11))) +
        (a[12] + a[13]) + (a[14] xor a[15]) + (a[16] xor (a[16] * 9)) +
        (a[1] + (a[1] xor x));

      n := (a[(x mod 16) + 1] mod modval) + lowr;
      Inc(x);


    until (was[n] = 0) and ((n <> i) or (i = highr));


    if (n = i) and (i = highr) then
      begin
      x := stable2[highr - 1];
      stable2[highr - 1] := n;
      bstable2[n] := highr - 1;
      n := x;
      end;

    was[n] := 1;

    stable2[i]  := n;
    bstable2[n] := i;

    end;

end;


function ffirst (s: string): integer;
var
  n: integer;
begin
  for n := 1 to length(s) do
    begin
    Result := n;
    if byte(s[n]) in [lowr..highr] then
      break;

    end;
end;


function k1crypt (s: string): string;
var
  i, idx, First: integer;
  nc, prv: byte;

begin
  Result := '';

  First := ffirst(s);
  idx := 0;
  for I := 1 to length(s) do
    begin

    if idx < 16 then
      Inc(idx)
    else
      idx := 1;


    if byte(s[i]) in [lowr..highr] then
      begin
      if i = First then
        prv := iv;
      nc := stable[((stable2[((byte(s[i]) xor prv) mod modval) + lowr] xor
        key[idx]) mod modval) + lowr];
      prv := nc;
      Result := Result + char(nc);
      end
    else
      Result := Result + s[i];

    end;

end;

function k1decrypt (s: string): string;
var
  i, idx, First: integer;
  nc, prv: byte;

  function fprv (index: integer): byte;
  begin
    repeat
      Result := byte(s[index]);
      Dec(index);
    until Result in [lowr..highr];
  end;

begin
  Result := '';

  First := ffirst(s);

  idx := 0;

  for I := 1 to length(s) do
    begin
    if idx < 16 then
      Inc(idx)
    else
      idx := 1;
    end;


  for I := length(s) downto 1 do
    begin
    if byte(s[i]) in [lowr..highr] then
      begin
      if i = First then
        prv := iv
      else
        prv := fprv(i - 1);
      nc := ((bstable2[(((bstable[(((byte(s[i]))))] - lowr) mod modval) xor
        key[idx])] - lowr) mod modval) xor prv;
      Result := char(nc) + Result;
      end
    else
      Result := s[i] + Result;


    if idx > 1 then
      Dec(idx)
    else
      idx := 16;

    end;

end;


function k2crypt (s: string): string;
var
  i: integer;
  x: longword;
  n, nc: byte;

var
  skey: array [1..16] of longword;

begin

  Result := '';

  x := 0;

  for I := 1 to 16 do
    begin
    skey[i] := key2[i];
    x := x + (skey[i] * i);
    end;


  for i := 1 to length(s) do
    begin

    Inc(x, (skey[1] + skey[2]) + ((skey[3] - skey[4]) xor
      (skey[5] shr (skey[6] mod 8))) + (skey[7] shl (skey[8] mod 8)) +
      (skey[6] xor skey[8]) xor ((skey[5] + i) xor (skey[7] - i)) +
      (skey[9] + skey[10] + ((skey[10] shl 6) xor (skey[10] shr 11))) +
      (skey[12] - skey[13]) + (skey[14] xor skey[15]) +
      (skey[16] xor (skey[16] * 9)) + (skey[1] + (skey[1] xor (i + 1))));

    skey[(i mod 16) + 1] := x;


    if byte(s[i]) in [lowr..highr] then
      begin

      n := (skey[(i mod 16) + 1] mod modval) + lowr;

      nc := ((byte(s[i]) xor n) mod modval) + lowr;

      end
    else
      begin
      nc := byte(s[i]);
      end;

    Result := Result + char(nc);

    end;
end;


function k2decrypt (s: string): string;
var
  i: integer;
  x: longword;
  n, nc: byte;

var
  skey: array [1..16] of longword;

begin

  Result := '';

  x := 0;

  for I := 1 to 16 do
    begin
    skey[i] := key2[i];
    x := x + (skey[i] * i);
    end;

  for i := 1 to length(s) do
    begin

    Inc(x, (skey[1] + skey[2]) + ((skey[3] - skey[4]) xor
      (skey[5] shr (skey[6] mod 8))) + (skey[7] shl (skey[8] mod 8)) +
      (skey[6] xor skey[8]) xor ((skey[5] + i) xor (skey[7] - i)) +
      (skey[9] + skey[10] + ((skey[10] shl 6) xor (skey[10] shr 11))) +
      (skey[12] - skey[13]) + (skey[14] xor skey[15]) +
      (skey[16] xor (skey[16] * 9)) + (skey[1] + (skey[1] xor (i + 1))));

    skey[(i mod 16) + 1] := x;


    if byte(s[i]) in [lowr..highr] then
      begin

      n := (skey[(i mod 16) + 1] mod modval) + lowr;

      nc := ((byte(s[i]) - lowr) mod modval) xor n;

      end
    else
      begin
      nc := byte(s[i]);
      end;

    Result := Result + char(nc);

    end;
end;


function crypt (s: string): string;
begin
  Result := k2crypt(k1crypt(s));
end;

function decrypt (s: string): string;
begin
  Result := k1decrypt(k2decrypt(s));
end;


end.
